"""Provides the Dataset related classes."""
from .base import Base


class Datasets(Base):
    _base_path_ = 'datasets'
    _auto_methods_ = ('get', )


class Tasks(Base):
    _base_path_ = 'tasks'
    _auto_methods_ = ('get', )


class Runs(Base):
    _base_path_ = 'runs'
    _auto_methods_ = ('get', )



"""Provides the User class."""
from .base import Base


class User(Base):
    _base_path_ = 'user'
    _auto_methods_ = ('get', 'post', 'put')

    def get_predictors(self, **kwargs):
        """ Get Predictors uploaded by user
        :return: client response object
        """
        return self.get(sub_route='predictors', **kwargs)
    
    def get_analyses(self, load=False):
        """ Get NeuroVault uploads associated with this analysis
        :param str id: Analysis hash_id.
        :return: client response object
        """
        
        return self.get(sub_route='myanalyses', **kwargs)



"""
    pyns
    ~~~~~~~~~~~~~~~~~~~
    A client for interfacing with http://neuroscout.org API
    :copyright: (c) 2018 by Alejandro de la Vega.
    :license: MIT, see LICENSE for more details.
"""

API_BASE_URL = 'https://neuroscout.org/api'
ROUTE_PATTERN = '{base_url}/{route}[/{id}][/{sub_route}]'

from .api import Neuroscout

__all__ = ['Neuroscout']

__author__ = ['Alejandro de la Vega']
__license__ = 'MIT'



import os
import pytest
import betamax
from betamax_serializers import pretty_json
from pathlib import Path

from pyns import Neuroscout

USER_TEST_EMAIL = os.environ.get('USER_TEST_EMAIL', '** secret user **')
USER_TEST_PWD = os.environ.get('USER_TEST_PWD', '** secret password **')

betamax.Betamax.register_serializer(pretty_json.PrettyJSONSerializer)

with betamax.Betamax.configure() as config:
    config.cassette_library_dir = 'tests/cassettes'
    config.default_cassette_options['serialize_with'] = 'prettyjson'
    config.define_cassette_placeholder('<USER_TEST_EMAIL>', USER_TEST_EMAIL)
    config.define_cassette_placeholder('<USER_TEST_PWD>', USER_TEST_PWD)


@pytest.fixture(scope='module')
def neuroscout_recorder():
    """ Sets up client, recorder pair """
    neuroscout = Neuroscout()
    return betamax.Betamax(neuroscout._session), neuroscout


@pytest.fixture(scope='module')
def recorder(neuroscout_recorder):
    """ Returns only recorder """
    return neuroscout_recorder[0]


@pytest.fixture(scope='module')
def neuroscout(neuroscout_recorder):
    """ Authorizes and returns client """
    recorder, neuroscout = neuroscout_recorder
    with recorder.use_cassette('auth'):
        neuroscout._authorize(email=USER_TEST_EMAIL, password=USER_TEST_PWD)
    return neuroscout


@pytest.fixture(scope='module')
def get_test_data_path():
    return Path(__file__).parents[0].absolute() / 'data'



from time import sleep


def test_predictor(recorder, neuroscout):
    with recorder.use_cassette('predictor'):
        resp = neuroscout.predictors.get(run_id=126)
        assert len(resp) > 20

        first = resp[0]
        assert 'source' in first

        resp = neuroscout.predictors.get(id=first['id'])
        assert 'source' in resp


def test_predictor_collection(recorder, neuroscout, get_test_data_path):
    with recorder.use_cassette('predictor_collection'):
        resp = neuroscout.predictors.create_collection(
                "test_collection", dataset_id=10, runs=[[10, 9], [8, 7]],
                event_files=[str(get_test_data_path / "new_events1.tsv"),
                             str(get_test_data_path / "new_events1.tsv")],
                descriptions={"reaction_time": "fake RT"})
        assert resp['collection_name'] == "test_collection"
        assert resp['status'] == "PENDING"

        new = neuroscout.predictors.get_collection(resp['id'])
        while new['status'] == "PENDING":
            sleep(1)
            new = neuroscout.predictors.get_collection(resp['id'])

        assert len(new['predictors']) == 3



"""Provide the Base superclass."""
from abc import ABC, abstractmethod
from functools import partial


class Base(ABC):
    """Superclass for all models."""

    def __init__(self, client):
        """Initialize a Model instance.
        :param client: An instance of :class:`.Neuroscout`.
        """
        self._client = client

        all_methods = ('get', 'post', 'put', 'delete')
        assert set(self._auto_methods_) <= set(all_methods)

        for method in self._auto_methods_:
            setattr(self,
                    method,
                    partial(
                        getattr(self._client, "_" + method),
                        self._base_path_)
                    )

    @property
    @abstractmethod
    def _base_path_(self):
        pass

    @property
    @abstractmethod
    def _auto_methods_(self):
        """ HTTP methods to auto create in subordinate classes """
        pass



from .user import User
from .analysis import Analyses
from .dataset import Datasets, Tasks, Runs
from .predictor import Predictors, PredictorEvents

__all__ = [
    'Analyses',
    'Datasets',
    'Tasks',
    'Runs',
    'Predictors',
    'PredictorEvents',
    'User'
]



"""Provides the Predictor related classes."""
from .base import Base
import json


class Predictors(Base):
    _base_path_ = 'predictors'
    _auto_methods_ = ('get', 'post')

    def create_collection(self, collection_name, dataset_id,
                          runs, event_files, descriptions=None):
        """ Create new predictor collection
        :param str collection_name: Force upload with unique timestamped name.
        :param int dataset_id: Dataset id.
        :param list(list((int)) runs: List of run ids corresponding to files
        :param list(str) event_files: TSV files with new predictor columns.
            Required columns: onset, duration,
            any number of columns with values for new Predictors.
        :param list(dict) descriptions: optional list of descriptions
                                        for each columns
        :return: JSON response
        """
        files = tuple([('event_files', open(f, 'rb')) for f in event_files])
        runs = [",".join([str(r) for r in s]) for s in runs]
        descriptions = json.dumps(descriptions)
        return self.post('collection', dataset_id=dataset_id, files=files,
                         runs=runs, collection_name=collection_name,
                         descriptions=descriptions)

    def get_collection(self, collection_id):
        return self.get(f'collection/{collection_id}')


class PredictorEvents(Base):
    _base_path_ = 'predictor-events'
    _auto_methods_ = ('get', )



def test_auth(recorder, neuroscout):
    assert neuroscout._api_token is not None
    assert len(neuroscout._api_token)


def test_datasets(recorder, neuroscout):
    with recorder.use_cassette('get_dataset'):
        resp = neuroscout._get('datasets')

        datasets = resp
        assert len(datasets) > 1



import requests
import re
import jwt
import os
from datetime import datetime
from functools import partialmethod

from . import API_BASE_URL, ROUTE_PATTERN
from . import models


class Neuroscout(object):
    """Neuroscout API client object. This is the access point for the API.

    Args:
        email (str, optional): Email address to use for authorization.
        password (str, optional): Password for authorization (not saved)
        api_base_url (str, optional): Alternate base URL for API

    """
    def __init__(self, email=None, password=None, api_base_url=None):
        self._session = requests.Session()
        self._api_base_url = api_base_url or API_BASE_URL
        self._api_token = None

        self._authorize(email, password)

        # Set up main routes
        self.analyses = models.Analyses(self)
        self.datasets = models.Datasets(self)
        self.tasks = models.Tasks(self)
        self.runs = models.Runs(self)
        self.predictors = models.Predictors(self)
        self.predictor_events = models.PredictorEvents(self)
        self.datasets = models.Datasets(self)
        self.user = models.User(self)

    def _get_headers(self):
        """ Build authorization header """
        if self._api_token is not None:
            return {'Authorization': 'JWT %s' % self._api_token}
        else:
            return None

    def _build_path(self, route, **kwargs):
        """ Build and format a URI for a request.

        Args:
            route (str): Primary API route. E.g. 'user'
            kwargs (dict): Dictionary of variables used to format URI.

        Returns:
            path (str): Formatted URI
        """
        def _replace_variables(pattern, variables):
            for name in re.findall("\{(.*?)\}", pattern):
                if name in variables and variables[name] is not None:
                    di = {name: str(variables[name])}
                    pattern = pattern.format(**di)
                else:
                    return ''

            return pattern

        new_path = ROUTE_PATTERN
        optional_patterns = re.findall('\[(.*?)\]', ROUTE_PATTERN)

        for pattern in optional_patterns:
            chunk = _replace_variables(pattern, kwargs)
            new_path = new_path.replace('[%s]' % pattern, chunk)

        return new_path.format(base_url=self._api_base_url, route=route)

    def _make_request(self, request, route, sub_route=None, id=None,
                      params=None, data=None,
                      json=None, headers=None, remove_null=True,
                      files=None, **kwargs):
        """ Generic request handler """

        if route != 'auth':
            self._check_expiry()

        request_function = getattr(self._session, request)

        if remove_null is True:
            kwargs = {k: v for (k, v) in kwargs.items() if v is not None}

        if request == 'get':
            params = kwargs
            # Join lists as comma separated list
            for k, v in params.items():
                if isinstance(v, list):
                    v = [str(i) for i in v]
                    params[k] = ','.join(v)

        elif request in ['put', 'post']:
            if files:
                data = kwargs
            else:
                json = kwargs

        headers = headers or self._get_headers()
        route = self._build_path(route, sub_route=sub_route, id=id)

        resp = request_function(
            route, json=json, data=data, files=files,
            headers=headers, params=params)

        if resp.headers.get('Content-Type') == 'application/json':
            content = resp.json()
        else:
            content = resp.content

        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if isinstance(content, dict):
                if content.get('message') is not None:
                    error = str(error) + "\n Message: {}".format(
                        content['message'])

            raise requests.exceptions.HTTPError(error)

        return content

    def _authorize(self, email=None, password=None):
        """ Fetch api_token given access credentials """
        if email is None and 'NEUROSCOUT_USER' in os.environ:
            email = os.environ['NEUROSCOUT_USER']
        if password is None and 'NEUROSCOUT_PASSWORD' in os.environ:
            password = os.environ['NEUROSCOUT_PASSWORD']

        if email is not None and password is not None:
            rv = self._post('auth', email=email, password=password)
            self._api_token = rv['access_token']

            iat = datetime.utcfromtimestamp(
                jwt.decode(self._api_token, verify=False)['iat'])
            adj = iat - datetime.now()
            self._api_token_exp = datetime.utcfromtimestamp(
                jwt.decode(self._api_token, verify=False)['exp']) - adj

    def _check_expiry(self):
        if self._api_token is not None:
            if datetime.now() > self._api_token_exp:
                self._authorize()

    _get = partialmethod(_make_request, 'get')
    _post = partialmethod(_make_request, 'post')
    _put = partialmethod(_make_request, 'put')
    _delete = partialmethod(_make_request, 'delete')



def test_dataset(recorder, neuroscout):
    with recorder.use_cassette('dataset'):
        resp = neuroscout.datasets.get()

        assert len(resp) > 1

        resp = neuroscout.datasets.get(id=5)
        runs = resp['runs']
        assert len(runs) > 10
        tasks = resp['tasks']
        assert len(tasks) == 2

        resp = neuroscout.tasks.get(dataset_id=5)
        assert resp[0]['name'] == tasks[0]['name']

        resp = neuroscout.runs.get(dataset_id=5)
        assert len(resp) == len(runs)



import pytest
from time import sleep
from requests.exceptions import HTTPError

TEST_ANALYSIS = {
  "dataset_id": 9,
  "model": {
    "Steps": [
      {
        "DummyContrasts": {"Type": "t"},
        "Contrasts": [],
        "Level": "Run",
        "Model": {
          "X": [
            "speech"
          ]
        },
        "Transformations": []
      },
      {
        "DummyContrasts": {
          "Type": "FEMA"
        },
        "Level": "Subject"
      },
      {
        "DummyContrasts": {"Type": "t"},
        "Level": "Dataset"
      }
    ],
    "Input": {
      "Run": [1],
      "Subject": [
        "rid000001"
      ],
      "Task": "life"
    },
    "Name": "pytest_analysis"
  },
  "name": "pytest_analysis",
  "predictors": [12905],
  "runs": [236],
  "task_name": "life",
}


@pytest.fixture(scope='module')
def analysis(recorder, neuroscout):
    """ Creates analysis """
    with recorder.use_cassette('analysis'):
        new = neuroscout.analyses.post(**TEST_ANALYSIS)

    return new


@pytest.fixture(scope='module')
def analysis_object(recorder, neuroscout):
    """ Create analysis object """
    with recorder.use_cassette('analysis_object'):
        new = neuroscout.analyses.create_analysis(
            name='pytest_analysis',
            dataset_name='Life',
            task='life',
            predictor_names=['speech'],
            hrf_variables=['speech'],
            subject=['rid000001'])

    return new


def test_analysis_object(recorder, neuroscout, analysis_object):
    assert analysis_object.status == 'DRAFT'
    assert analysis_object.name == 'pytest_analysis'
    assert analysis_object.description is None

    analysis_object.description = 'new_description'

    with recorder.use_cassette('test_analysis_object'):
        analysis_object.push()
        assert analysis_object.description == 'new_description'

        analysis_object.description = 'overwritten'
        analysis_object.pull()
        assert analysis_object.description == 'new_description'

        assert analysis_object.get_status()['status'] == 'DRAFT'

        # test fill
        analysis_object.predictors = []
        analysis_object.push()
        analysis_object.fill()
        assert len(analysis_object.predictors) == 1

        # Test get_report
        resp = analysis_object.generate_report(run_id=analysis_object.runs[0])
        assert resp['status'] == 'PENDING'

        resp = analysis_object.get_report(run_id=analysis_object.runs[0])

        assert len(resp['result']['design_matrix']) == 1

        # compile
        analysis_object.compile(build=False)
        assert analysis_object.get_status()['status'] != 'DRAFT'

        assert 'dataset_address' in analysis_object.get_resources()
        assert hasattr(analysis_object, 'dataset_address')

        while analysis_object.get_status()['status'] != 'PASSED':
            sleep(2)

        # clone
        new = analysis_object.clone()
        assert new.name == analysis_object.name
        assert new.hash_id != analysis_object.hash_id
        assert new.parent_id == analysis_object.hash_id

        # Create object from existing id
        same = neuroscout.analyses.get_analysis(id=new.hash_id)
        assert same.hash_id == new.hash_id
        assert same.name == new.name

        # Test delete
        new.delete()

        with pytest.raises(HTTPError):
            analysis_object._analyses.get(id=new.hash_id)


def test_get_analysis(recorder, neuroscout, analysis):
    analysis_id = analysis['hash_id']

    # Test get
    with recorder.use_cassette('get_analysis'):
        resp = neuroscout.analyses.get(id=analysis_id)
        assert resp['name'] == 'pytest_analysis'
        assert resp['status'] == 'DRAFT'


def test_put_analysis(recorder, neuroscout, analysis):
    analysis_id = analysis['hash_id']

    analysis['description'] = 'new_description'

    # Test put
    with recorder.use_cassette('put_analysis'):
        resp = neuroscout.analyses.put(id=analysis_id, **analysis)

        assert resp['name'] == 'pytest_analysis'
        assert resp['description'] == 'new_description'


def test_id_actions(recorder, neuroscout, analysis):
    analysis_id = analysis['hash_id']

    with recorder.use_cassette('id_analysis'):
        # Test full
        resp = neuroscout.analyses.get_full(id=analysis_id)
        assert 'runs' in resp

        # Test get_report
        resp = neuroscout.analyses.generate_report(id=analysis_id,
                                                   run_id=analysis['runs'][0])
        assert resp['status'] == 'PENDING'

        # Wait until compiled
        resp = neuroscout.analyses.get_report(id=analysis_id,
                                              run_id=analysis['runs'][0])

        assert len(resp['result']['design_matrix']) == 1

        # Test fill

        analysis['runs'] = []
        analysis['predictors'] = []

        # Put first
        resp = neuroscout.analyses.put(id=analysis_id, **analysis)
        assert len(resp['predictors']) == 0

        resp = neuroscout.analyses.fill(
            id=analysis_id)

        assert len(resp['runs']) == 76
        assert len(resp['predictors']) == 1

        # Test compile
        resp = neuroscout.analyses.compile(id=analysis_id, build=False)
        assert resp['status'] == 'PENDING'

        # Test status
        resp = neuroscout.analyses.get_status(id=analysis_id)
        assert 'status' in resp

        # Wait until compiled
        while(resp['status'] == 'PENDING'):
            sleep(1)
            resp = neuroscout.analyses.get_status(id=analysis_id)

        # Test resources
        resp = neuroscout.analyses.get_resources(id=analysis_id)
        assert 'dataset_address' in resp

        # Test bundle
        resp = neuroscout.analyses.get_bundle(id=analysis_id)

        # Test clone
        resp = neuroscout.analyses.clone(id=analysis_id)
        new_id = resp['hash_id']
        assert new_id != analysis_id

        # Test delete
        resp = neuroscout.analyses.delete(id=new_id)


def test_upload_analysis(recorder, neuroscout, analysis, get_test_data_path):
    analysis_id = '9hvm3'  # This is an existing analysis on the api

    with recorder.use_cassette('upload_analysis'):
        group_paths = [str(p) for p in
                       (get_test_data_path / 'fitlins').glob('task*')]
        sub_paths = [str(p) for p in
                     (get_test_data_path / 'fitlins').glob('sub*')]
        resp = neuroscout.analyses.upload_neurovault(
            id=analysis_id, group_paths=group_paths,
            subject_paths=sub_paths,
            validation_hash='Av1Jbwl1aO', force=True,
            n_subjects=99)

        assert 'uploaded_at' in resp
        uploaded_at = resp['uploaded_at']

        # Check uploads
        uploads = neuroscout.analyses.get_uploads(id=analysis_id)

        assert len(uploads) >= 1
        newest = [u for u in uploads if u['uploaded_at'] == uploaded_at][0]

        # Wait until compiled
        while(
          any([True for f in newest['files'] if f['status'] == 'PENDING'])):
            sleep(2)
            uploads = neuroscout.analyses.get_uploads(id=analysis_id)
            newest = [u for u in uploads if u['uploaded_at'] == uploaded_at]
            if len(newest) > 0:
                newest = newest[0]

        assert all([f['status'] == 'OK' for f in newest['files']])
        assert newest['collection_id'] is not None


def test_load_analysis(recorder, neuroscout):
    with recorder.use_cassette('load_analysis'):
      m = neuroscout.analyses.get_analysis('gbp6i')
      assert len(m.load_uploads()) == 3
      assert len(m.load_uploads(select=None)) == 3
      nistats = m.load_uploads(
        collection_filters=dict(estimator='nistats'))
      latest_up = nistats[0][1]['uploaded_at']
      assert len(nistats) == 3
      assert nistats[0][1]['estimator'] == 'nistats'



from setuptools import setup, find_packages
PACKAGES = find_packages()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='pyns',
      version='0.4.10',
      description='Neuroscout API wrapper',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://github.com/neuroscout/pyns',
      author='Alejandro de la Vega',
      author_email='aleph4@gmail.com',
      install_requires=['requests>=2.21', 'pyjwt~=1.7.1', 'tqdm>=4.30.0'],
      license='MIT',
      packages=PACKAGES,
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      zip_safe=False)



"""Provides the Analysis related classes."""
from .base import Base
from pathlib import Path
from functools import partial
import datetime
import tempfile
import requests
from .utils import build_model, attempt_to_import
import tqdm
import time
import re
import json

altair = attempt_to_import('altair')
nib = attempt_to_import('nibabel')
nilearn = attempt_to_import('nilearn.plotting')


TMP_DIR = Path(tempfile.mkdtemp())


class Analysis:
    """ Analysis object class. Object representing an analysis that can be
    synced with the API """

    _mutable_fields_ = ['dataset_id', 'description', 'name',  'predictions',
                        'model', 'predictors', 'private', 'runs']

    _aliased_methods_ = ['delete', 'get_bundle', 'compile', 'generate_report',
                         'get_report', 'upload_neurovault', 'get_uploads',
                         'load_uploads', 'plot_uploads',
                         'plot_report', 'get_design_matrix']

    def __init__(self, *, analyses, name, dataset_id, **kwargs):
        """ Initate a new Analysis object. Typically, this is done by
        'get_analysis' or 'create_analysis'.
        Args:
            self (obj)
            analyses (obj)- Instantiated analyses object
            name (str) - Analysis name
            dataset_id (int) - ID of dataset
        """
        self.name = name
        self.dataset_id = dataset_id
        self._analyses = analyses

        # Set up (invalid fields will also be set, but not pushed to API)
        for k, v in kwargs.items():
            setattr(self, k, v)

        # If no hash_id, create
        if not hasattr(self, 'hash_id'):
            self._fromdict(self._analyses.post(**self._asdict()))

        # Attach aliased methods
        for method in self._aliased_methods_:
            setattr(self,
                    method,
                    partial(
                        getattr(self._analyses, method),
                        self.hash_id)
                    )

    def __repr__(self):
        return "<Analysis hash_id={} name={} dataset_id={}>".format(
            self.hash_id, self.name, self.dataset_id)

    def _asdict(self):
        """ Return dictionary representation of mutable fields """
        di = {}
        for field in self._mutable_fields_:
            if hasattr(self, field):
                di[field] = getattr(self, field)

        return di

    def _fromdict(self, di):
        """ Update field values from response """
        for k, v in di.items():
            setattr(self, k, v)

    def push(self):
        """ Push changes from to API, and sync with returned results"""
        self._fromdict(self._analyses.put(self.hash_id, **self._asdict()))

    def pull(self):
        """ Pull updates from API, overriding changes made locally """
        self._fromdict(self._analyses.get(self.hash_id))

    def _getter_wrapper(self, method, **kwargs):
        """ Get representation of analysis, sync and return """
        new = getattr(self._analyses, method)(self.hash_id, **kwargs)
        self._fromdict(new)
        return new

    def fill(self, partial=True, dryrun=False):
        """ Fill missing fields from API
        :param bool partial: Partial fill?
        :param bool dryrun: Dryrun do not commit to database.
        :return: client response object
        """
        return self._getter_wrapper('fill')

    def get_status(self):
        """ Get compilation status """
        return self._getter_wrapper('get_status')

    def get_resources(self):
        """ Get analysis resources """
        return self._getter_wrapper('get_resources')

    def get_full(self):
        """ Get full analysis representation """
        return self._getter_wrapper('get_full')

    def clone(self, dataset_id=None):
        """ Clone current analysis, and return a new Analysis object
        :param int dataset_id: If dataset_id is provided, new run and
                               predictor_ids will be filled for that dataset.
        """
        new = Analysis(
            analyses=self._analyses, **self._analyses.clone(self.hash_id))
        if dataset_id is not None:
            new.dataset_id = dataset_id
            new.runs = []
            new.predictors = []
            new.push()
            new.fill()
        return new


class Analyses(Base):
    """ Class used to access analysis API route """
    _base_path_ = 'analyses'
    _auto_methods_ = ('get', 'post')

    def put(self, id, **kwargs):
        """ Put analysis
        :param str id: Analysis hash_id.
        :return: client response object
        """
        return self._client._put(self._base_path_, id=id, **kwargs)

    def delete(self, id):
        """ Delete analysis
        :param str id: Analysis hash_id.
        :return: client response object
        """
        return self._client._delete(self._base_path_, id=id)

    def get_bundle(self, id, filename=None):
        """ Get analysis bundle
        :param str id: Analysis hash_id.
        :param str, object filename: Optional filename to save bundle to
        :return: client response object
        """
        bundle = self.get(id=id, sub_route='bundle')
        if filename is not None:
            if isinstance(filename, str):
                filename = Path(filename)
            with filename.open('wb') as f:
                f.write(bundle)
        else:
            return bundle

    def clone(self, id):
        """ Clone analysis
        :param str id: Analysis hash_id.
        :return: client response object, with new analysis id
        """
        return self.post(id=id, sub_route='clone')

    def create_analysis(self, *, name, dataset_name, predictor_names,
                        tasks=None, subjects=None, runs=None, session=None,
                        hrf_variables=None, contrasts=None,
                        dummy_contrasts=True, transformations=None, **kwargs):
        """ Analysis creation "wizard". Given run selection filters, and name
        of Predictors, builds Analysis object with prepopulated BIDS model.
        """

        # Get dataset id
        datasets = self._client.datasets.get(active_only=False)
        dataset = [d for d in datasets if d['name'] == dataset_name]
        if len(dataset) != 1:
            raise ValueError(
                "Dataset name does not match any existing dataset.")
        else:
            dataset = dataset[0]

        # Get task name
        if tasks is not None:
            if not isinstance(tasks, list):
                tasks = [tasks]

            task_ids = []
            for task in tasks:
                search = [t for t in dataset['tasks'] if t['name'] == task]
                if len(search) != 1:
                    raise ValueError(
                        "Task name does not match any tasks in the dataset")
                task_ids.append(search[0]['id'])
        else:
            # All tasks
            tasks = [t['name'] for t in dataset['tasks']]
            task_ids = [t['id'] for t in dataset['tasks']]

        # Get Run IDs
        run_models = self._client.runs.get(
            dataset_id=dataset['id'], task_id=task_ids,
            subject=subjects, number=runs, session=session)

        if len(run_models) < 1:
            raise ValueError("No runs could be found with the given criterion")

        subjects = list(set(r['subject'] for r in run_models))
        runs = list(set(r['number'] for r in run_models if r['number']))
        runs = runs or None

        run_ids = [r['id'] for r in run_models]
        # Get Predictor IDs
        public_preds = self._client.predictors.get(
            run_id=run_ids, name=predictor_names, active_only=False)

        predictors = [p['id'] for p in public_preds]

        # If not all predictors found, search in user private predictors
        private_preds = set(predictor_names) - \
            set([p['name'] for p in public_preds])
        if private_preds:
            # Get Predictor IDs
            for pred in private_preds:
                predictors += [p['id']
                               for p in self._client.user.get_predictors(
                                   run_id=run_ids, name=pred)]

        if len(predictors) != len(predictor_names):
            raise ValueError(
                "Not all named predictors could be found for the "
                "specified runs.")

        # Build model
        if transformations:
            transformations = transformations.copy()
        model = build_model(
            name, predictor_names, tasks,
            subjects=subjects, runs=runs, session=session,
            hrf_variables=hrf_variables,
            transformations=transformations,
            contrasts=contrasts, dummy_contrasts=dummy_contrasts
            )

        analysis = Analysis(analyses=self, dataset_id=dataset['id'],
                            name=name, model=model, runs=run_ids,
                            predictors=predictors, **kwargs)

        return analysis

    def get_analysis(self, id):
        """ Convenience function to fetch and create Analysis object from
        a known analysis id
        :param str id: Analysis hash_id.
        :return: Analysis object
        """
        return Analysis(analyses=self, **self.get(id=id))

    def compile(self, id,  build=True):
        """ Submit analysis for complilation
        :param str id: Analysis hash_id.
        :param bool build: Build pybids object and verify compilation
        :return: client response object
        """
        return self.post(id=id, sub_route='compile', params=dict(build=build))

    def generate_report(self, id, run_id=None, sampling_rate=None, scale=False):
        """ Submit analysis for report generation
        :param str id: Analysis hash_id.
        :param int run_id: Optional run_id to constrain report.
        :return: client response object
        """
        return self.post(id=id, sub_route='report', 
                         params=dict(
                             run_id=run_id, sampling_rate=sampling_rate, scale=scale))

    def get_report(self, id, run_id=None, loop_wait=True):
        """ Get generated reports for analysis
        :param str id: Analysis hash_id.
        :param int run_id: Optional run_id to constrain report.
        :return: client response object
        """
        report = self.get(id=id, sub_route='report', run_id=run_id)
        if loop_wait:
            while report['status'] == 'PENDING':
                time.sleep(2)
                report = self.get(id=id, sub_route='report', run_id=run_id)

        return report

    def get_design_matrix(self, id, run_id=None,
                          loop_wait=True):
        """ Get report design_matrix
        :param str id: Analysis hash_id
        :param int run_id: Optional run_id to constrain report.
        :param boolean loop_wait: Wait for result from report
        """
        report = self.get_report(id=id, run_id=run_id, loop_wait=loop_wait)

        if report['status'] == 'OK':
            return report['result']['design_matrix']
        else:
            return None

    def plot_report(self, id, run_id=None, plot_type='design_matrix_plot',
                    loop_wait=True):
        """ Uses altair to plot design_matrix plot generated by report
        :param str id: Analysis hash_id
        :param int run_id: Optional run_id to constrain report.
        :param str plot_type: Plot type to display
        :param boolean loop_wait: Wait for result from report
        """
        if altair is None:
            raise ImportError("Altair is required to plot_reports")

        report = self.get_report(id=id, run_id=run_id, loop_wait=loop_wait)

        if report['status'] == 'OK':
            for p in report['result'][plot_type]:
                altair.display.vegalite(p)

        return None

    def upload_neurovault(self, id, validation_hash, subject_paths=None,
                          group_paths=None, collection_id=None, force=False,
                          cli_version=None, fmriprep_version=None,
                          estimator=None, n_subjects=None, cli_args=None):
        """ Submit analysis for report generation
        :param str id: Analysis hash_id.
        :param str validation_hash: Validation hash string.
        :param list(str) subject_paths: List of image paths.
        :param list(str) group_paths: List of image paths.
        :param bool force: Force upload with unique timestamped name.
        :param str: neuroscout-cli version at runtime
        :param str: fmriprep version at runtime
        :param str: estimator used in fitlins (anfi/nilearn)
        :param int n_subjects: Number of subjects in analysis.
        :param dict runtime_options 
        :return: dict cli_args Arguments specified to CLI at runtime
        """

        def _ts_first(paths):
            tmaps = [t for t in paths if 'stat-t' in t]
            for t in tmaps:
                paths.remove(t)

            return tmaps + paths

        req = None
        # Do group, then subject level
        if group_paths is not None:
            print("Uploading group images")
            for path in tqdm.tqdm(_ts_first(group_paths)):
                files = {'image_file': open(path, 'rb')}
                req = self.post(
                    id=id, sub_route='upload', files=files, level='GROUP',
                    validation_hash=validation_hash, force=force,
                    fmriprep_version=fmriprep_version, estimator=estimator,
                    cli_version=cli_version, n_subjects=n_subjects, 
                    cli_args=json.dumps(cli_args), collection_id=collection_id)
                if collection_id is None:
                    collection_id = req['collection_id']

        if subject_paths is not None:
            print("Uploading subject images")
            for path in tqdm.tqdm(_ts_first(subject_paths)):
                files = {'image_file': open(path, 'rb')}
                req = self.post(
                    id=id, sub_route='upload', files=files, level='SUBJECT',
                    validation_hash=validation_hash, force=force,
                    fmriprep_version=fmriprep_version, estimator=estimator,
                    cli_version=cli_version, collection_id=collection_id)
                if collection_id is None:
                    collection_id = req['collection_id']

        if req is None:
            print("No images found")

        return req

    def get_uploads(self, id, select='latest', **kwargs):
        """ Get NeuroVault uploads associated with this analysis
        :param str id: Analysis hash_id.
        :param str select: How to select from multiple collections.
        Options: "latest", "oldest" or None. If None, returns all results.
        :param dict kwargs: Attributes to filter collections on.
         If any attributes are not found, they are ignored.
        :return: client response object
        """
        uploads = self.get(id=id, sub_route='upload')
        
        # Sort by date
        uploads = sorted(uploads, key=lambda x: datetime.datetime.strptime(
                x['uploaded_at'], '%Y-%m-%dT%H:%M'),
                         reverse=(select == 'latest'))

        # Select collections based on filters
        uploads = [
            u for u in uploads
            if all([u.get(k) == v for k, v in kwargs.items() if k in u])
            ]
        
        if not uploads:
            return None

        # Select first item unless all are requested
        if select is not None:
            uploads = [uploads[0]]

        return uploads

    def load_uploads(self, id, select='latest',
                     download_dir=None, collection_filters={}, 
                     image_filters={}):
        """ Load collection upload as NiBabel images and associated meta-data
        You can filter which images are loaded based on either collection
        level attributes or statmap image level attributes. These correspond
        to field returns for `get_uploads` at the collection level or
        `file` level. In addition for images, BIDS entities are parsed
        and available to filter on.

        :param str id: Analysis hash_id.
        :param str select: How to select from multiple collections.
        Options: "latest", "oldest" or None. If None, returns all results.
        :param str download_dir: Path to download images. If None, tempdir.
        :param dict collection_filters: Attributes to filter collections on.
        :param dict image_filters: Attributes to filter images on.
        If any attributes are not found, they are ignored.
        :return list list of tuples of format (Nifti1Image, kwargs).
        """
        if download_dir is None:
            download_dir = TMP_DIR
        else:
            download_dir = Path(download_dir)

        # Sort uploads for upload date
        uploads = self.get_uploads(id, **collection_filters)
        
        if not uploads:
            return None

        # Extract entities from file path
        def _get_entities(path):
            di = {}
            for t in ['task', 'contrast', 'stat', 'space']:
                matches = re.findall(f"{t}-(.*?)_", path)
                if matches:
                    di[t] = matches[0]
            return di

        # Filter files, download if necessary and load as Niimg-object
        flat = []
        for u in uploads:
            for f in u.pop('files'):
                f = {**f, **_get_entities(f['basename'])}

                # If file matches kwargs and is in NV
                if f.pop('status') == 'OK' and all(
                  [f.get(k, None) == v  if k in f else False for k, v in image_filters.items()]):
                    # Download and open
                    img_url = "https://neurovault.org/media/images/" \
                        f"{u['collection_id']}/{f['basename']}"
                    f_name = download_dir / \
                        f"{u['collection_id']}_{f['basename']}"

                    if not f_name.exists():
                        print(".", end ="")  
                        with f_name.open('wb') as file:
                            file.write(requests.get(img_url).content)
                    niimg = nib.load(f_name)

                    f.pop('traceback')
                    flat.append((niimg, {**u, **f}))
        return flat

    def plot_uploads(self, id, plot_args={}, **kwargs):
        """ Plot uploads for matching collections using nilearn
        :param str id: Analysis hash_id.
        :param dict kwargs: Arguments for load_uploads
        :return list list of matplotlib objects.
        """

        images = self.load_uploads(id, **kwargs)

        if images:
            plots = []
            for niimg, _ in images:
                plots.append(
                    nilearn.plotting.plot_stat_map(niimg, **plot_args))

            return plots
        else:
            return None

    def get_full(self, id):
        """ Get full analysis object (including runs and predictors)
        :param str id: Analysis hash_id.
        :return: client response object
        """
        return self.get(id=id, sub_route='full')

    def fill(self, id, partial=True, dryrun=False):
        """ Fill missing fields
        :param str id: Analysis hash_id.
        :param bool partial: Partial fill?
        :param bool dryrun: Dryrun do not commit to database.
        :return: client response object
        """
        return self.post(id=id, sub_route='fill',
                         params=dict(partial=partial, dryrun=dryrun))

    def get_resources(self, id):
        """ Get analysis resources
        :param str id: Analysis hash_id.
        :return: client response object
        """
        return self.get(id=id, sub_route='resources')

    def get_status(self, id):
        """ Get analysis status
        :param str id: Analysis hash_id.
        :return: client response object
        """
        return self.get(id=id, sub_route='compile')



""" Miscelaneous utilities """
import collections

module_names = {}
Dependency = collections.namedtuple('Dependency', 'package value')


def attempt_to_import(dependency, name=None, fromlist=None):
    if name is None:
        name = dependency
    try:
        mod = __import__(dependency, fromlist=fromlist)
    except ImportError:
        mod = None
    module_names[name] = Dependency(dependency, mod)
    return mod


def build_model(name, variables, tasks, subjects, runs=None, session=None,
                hrf_variables=None, transformations=None,
                contrasts=None, dummy_contrasts=True):
    """ Builds a basic two level BIDS-Model """
    hrf_variables = hrf_variables or []
    transformations = transformations or []
    contrasts = contrasts or []

    if not set(variables) >= set(hrf_variables):
        raise ValueError("HRF Variables must be a subset of all variables")

    if hrf_variables:
        transformations.append({
            "Input": hrf_variables,
            "Name": "Convolve"
        })

    model = {
        "Steps": [
          {
            "DummyContrasts": {"Type": "t"},
            "Contrasts": contrasts,
            "Level": "Run",
            "Model": {
              "X": variables
            },
            "Transformations": transformations
          }
        ],
        "Input": {
          "Subject": subjects,
          "Task": tasks
        },
        "Name": name,
    }

    model['Steps'].append(
        {
            "DummyContrasts": {"Type": "FEMA"},
            "Level": "Subject"
        }
    )

    model['Steps'].append(
        {
            "DummyContrasts": {"Type": "t"},
            "Level": "Dataset"
        }
    )

    if not dummy_contrasts:
        model['Steps'][0].pop('DummyContrasts')

    if dummy_contrasts == 'hrf' and hrf_variables:
        model['Steps'][0]['DummyContrasts']['Conditions'] = hrf_variables

    if runs is not None:
        model['Input']['Run'] = runs

    if session is not None:
        model['Input']['Session'] = session

    return model


def snake_to_camel(string):
    words = string.split('_')
    return words[0] + ''.join(word.title() for word in words[1:])